import math


class Vector:
    """
        Vector class
        Supported operators:
            +: add two vectors
            -: subtract two vectors
            *: scalar multiplication
            /: scalar division
            ==: check if two vectors are equal
    """

    def __init__(self, **kwargs):
        """Initialize a vector with x, y and z coordinates.

            Params:
                x: x coordinate
                y: y coordinate
                z: z coordinate
                dimension: dimension of the vector. Default is 3
            Returns:
                A vector class instance.
            Raises:
                None
        """
        if kwargs is None:
            self.null()
        else:
            self.axes = ['x', 'y', 'z']
            self.dimension = kwargs.get('dimension', 3)
            self.check_axes(kwargs)

    def check_axes(self, kwargs: dict) -> None:
        """
            Check if the axes are valid.

            Params:
                kwargs: dictionary with the axes.

            Returns:
                None

            Raises:
                None
        """
        for axis in self.axes:
            if axis in kwargs:
                setattr(self, axis, kwargs[axis])
            else:
                setattr(self, axis, 0)

    def null(self) -> None:
        """
            Set all the coordinates to 0.

            Params:
                None

            Returns:
                None
        """
        for axis in self.axes[0: self.dimension]:
            setattr(self, axis, 0)

    def get(self, axis: str) -> float:
        """
            Get the value of the axis.

            Params:
                axis: axis to get the value.

            Returns:
                Value of the axis as a float.

            Raises:
                ValueError if the axis is not valid.
        """
        if axis not in self.axes:
            raise ValueError('Axis must be x, y or z')
        return getattr(self, axis)

    def magnitude(self) -> float:
        """
            Calculate the magnitude of the vector.

            Params:
                None

            Returns:
                Magnitude of the vector as a float.

            Raises:
                None
        """
        _sum = 0
        for axis in self.axes[0: self.dimension]:
            _sum += getattr(self, axis) ** 2
        return math.sqrt(_sum) if _sum > 0 else 0

    def midpoint(self) -> 'Vector':
        """
            Calculate the midpoint of the vector.

            Params:
                None

            Returns:
                Midpoint of the vector as a vector.

            Raises:
                ValueError if the vector is null.
        """
        if self.isnull():
            raise ValueError('Cannot calculate midpoint with null vector')
        return self / 2

    def opposite(self) -> 'Vector':
        """
            Return the opposite vector.

            Params:
                None

            Returns:
                A vector with the opposite coordinates of the original vector.

            Raises:
                ValueError if the vector is null.
        """
        if self.isnull():
            raise ValueError('Cannot calculate opposite with null vector')
        x, y, z = -self.get('x'), -self.get('y'), -self.get('z')
        return Vector(x=x, y=y, z=z, dimension=self.dimension)

    def isnull(self) -> bool:
        """
            Check if the vector is null.

            Params:
                None

            Returns:
                True if the vector is null, False otherwise.

            Raises:
                None
        """
        return self.magnitude() >= 0 and self.magnitude() == 0

    def direction_cosine(self, axis: str, degrees: bool = False, decimals: int = 2) -> float:
        """
            Calculate the direction cosine of the vector.

            Params:
                axis (str): axis to calculate the direction cosine.
                degrees (bool): if True, return the result in degrees.
                decimals (int): number of decimals to round the result.

            Returns:
                Direction cosine of the vector as a radian or degree.

            Raises:
                ValueError if the axis is not valid.
        """
        if axis not in self.axes:
            raise ValueError('Axis must be x, y or z')
        value = getattr(self, axis)
        radians = math.acos(value/self.magnitude())
        return math.degrees(radians).__round__(decimals) if degrees else radians

    def perpendicular(self, other: 'Vector') -> bool:
        """
            Check if the vector is perpendicular to the other vector.

            Params:
                other (Vector): other vector to check.

            Returns:
                True if the vector is perpendicular to the other vector, False otherwise.

            Raises:
                ValueError if the vector is null or the dimensions are not equal.
        """
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        if other.isnull() or self.isnull():
            raise ValueError('Cannot calculate perpendicular with null vector')
        return self * other == 0

    def angle(self, other: 'Vector', degrees: bool = False, decimals: int = 2) -> float:
        """
            Calculate the angle between the vectors.

            Params:
                other (Vector): other vector to calculate the angle.
                degrees (bool): if True, return the result in degrees.
                decimals (int): number of decimals to round the result.

            Returns:
                Angle between the vectors as a radian or degree.

            Raises:
                ValueError if the vector is null or the dimensions are not equal.
        """
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        if other.isnull() or self.isnull():
            raise ValueError('Cannot calculate angle with null vector')
        radians = math.acos(
            self * other / (self.magnitude() * other.magnitude()))
        return math.degrees(radians).__round__(decimals) if degrees else radians

    def projection(self, other: 'Vector') -> ['self->other', 'other->self']:
        """
            Calculate the projection of the vector on the other vector.

            Params:
                other (Vector): other vector to calculate the projection.

            Returns:
                A tuple with the projection:
                    - self->other: projection of the vector on the other vector.
                    - other->self: projection of the other vector on the vector.

            Raises:
                ValueError if the vector is null or the dimensions are not equal.
        """
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        if other.isnull() or self.isnull():
            raise ValueError('Cannot calculate proyection with null vector')
        product_scalar = self * other
        other_magnitude = other.magnitude() ** 2
        projection_magnitude = product_scalar / other_magnitude
        projection = other * projection_magnitude
        return projection, self - projection

    def cross(self, other: 'Vector') -> 'Vector':
        """
            Calculate the cross product with the other vector.

            Params:
                other (Vector): other vector to calculate the cross product.

            Returns:
                A vector with the cross product.

            Raises:
                ValueError if the vector is null or the dimensions are not equal.
        """
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        if other.isnull() or self.isnull():
            raise ValueError('Cannot calculate cross product with null vector')
        x = self.get('y') * other.get('z') - self.get('z') * other.get('y')
        y = self.get('x') * other.get('z') - self.get('z') * other.get('x')
        z = self.get('x') * other.get('y') - self.get('y') * other.get('x')
        return Vector(x=x, y=-y, z=z, dimension=self.dimension)

    def triple(self, other: 'Vector', third: 'Vector') -> float:
        """
            Calculate the triple product of the vectors.

            Params:
                other (Vector): other vector to calculate the triple product.
                third (Vector): third vector to calculate the triple product.

            Returns:
                Triple product of the vectors as a float.

            Raises:
                ValueError if the vector is null or the dimensions are not equal.
        """
        if self.dimension != other.dimension or self.dimension != third.dimension:
            raise ValueError('Dimensions must be equal')
        if other.isnull() or self.isnull() or third.isnull():
            raise ValueError(
                'Cannot calculate triple product with null vector')
        cross = other.cross(third)
        return self * cross

    def __eq__(self, other: 'Vector') -> bool:
        """
            Check if the vectors are equal.

            Params:
                other: other vector to check.

            Returns:
                True if the vectors are equal, False otherwise.

            Raises:
                ValueError if the dimensions are not equal.
        """
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        for axis in self.axes[0: self.dimension]:
            if getattr(self, axis) != getattr(other, axis):
                return False
        return True

    def __str__(self) -> str:
        """
            Return the vector as a string.

            Params:
                None

            Returns:
                Vector as a string in the form of (x, y, z).

            Raises:
                None
        """
        values = '('
        for axis in self.axes[0: self.dimension]:
            values += f'{getattr(self, axis)},'
        return values[:-1] + ')'

    def __repr__(self) -> str:
        """
            Return the vector as a string.

            Params:
                None

            Returns:
                Vector as a string in the form of Vector(x, y, z).

            Raises:
                None
        """
        return f'Vector({self.get("x")}, {self.get("y")}, {self.get("z")})'

    def __add__(self, other: 'Vector') -> 'Vector':
        """
            Add two vectors.

            Params:
                other: other vector to add.

            Returns:
                A vector with the sum of the two vectors.

            Raises:
                ValueError if the dimensions are not equal.
        """
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        vec = Vector(dimension=self.dimension)
        for axis in self.axes[0: self.dimension]:
            setattr(vec, axis, getattr(self, axis) + getattr(other, axis))
        return vec

    def __sub__(self, other: 'Vector') -> 'Vector':
        """
            Subtract two vectors.

            Params:
                other: other vector to subtract.

            Returns:
                A vector with the difference of the two vectors.

            Raises:
                ValueError if the dimensions are not equal.
        """
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        vec = Vector(dimension=self.dimension)
        for axis in self.axes[0: self.dimension]:
            setattr(vec, axis, getattr(self, axis) - getattr(other, axis))
        return vec

    def __mul__(self, other: 'Vector' or float) -> 'Vector':
        """
            Multiply a vector by a scalar or another vector.

            Params:
                other: scalar or vector to multiply.

            Returns:
                A vector with the product of the vector and the scalar or a vector.

            Raises:
                ValueError if the dimensions are not equal.
        """
        if isinstance(other, Vector):
            if self.dimension != other.dimension:
                raise ValueError('Dimensions must be equal')
            if other.isnull() or self.isnull():
                return x
            x = self.get('x') * other.get('x')
            y = self.get('y') * other.get('y')
            z = self.get('z') * other.get('z')
            return x + y + z
        vec = Vector(dimension=self.dimension)
        for axis in self.axes[0: self.dimension]:
            setattr(vec, axis, getattr(self, axis) * other)
        return vec

    def __truediv__(self, scalar: float) -> 'Vector':
        """
            Divide a vector by a scalar.

            Params:
                scalar: scalar to divide.

            Returns:
                A vector with the division of the vector by the scalar.
        """
        vec = Vector(dimension=self.dimension)
        for axis in self.axes[0: self.dimension]:
            setattr(vec, axis, getattr(self, axis) / scalar)
        return vec



from matplotlib import pyplot as plt
from .vector import Vector
from .point import Point


class Plot:
    """
        Plot class, uses matplotlib.
        Only supports 3d projection for now.
    """

    def __init__(self, **kwargs):
        """
            Initialize a Plot object.

            Params:
                name (str): name of the plot
                projection (str): projection of the plot. Default is 3d
                range list(float): range of the plot. Default is [0, 5]

            Returns:
                A plot class instance.

            Raises:
                ValueError if no arguments given
        """
        if kwargs is None:
            raise ValueError('No arguments given.')
        self.name = kwargs.get('name', 'Plot')
        self.projection = kwargs.get('projection', '3d')
        self.range = kwargs.get('range', [0, 5])
        self.prepare()

    def prepare(self):
        """
            Prepare the plot.

            Params:
                None

            Returns:
                None

            Raises:
                None
        """
        self.ax = plt.axes(projection=self.projection)
        self.ax.set_title(self.name)
        self.ax.set_xlabel('x')
        self.ax.set_ylabel('y')
        self.ax.set_zlabel('z')
        self.ax.set_xlim(self.range)
        self.ax.set_ylim(self.range)
        self.ax.set_zlim(self.range)

    def add_vector(self, vector: Vector, origin: Point, color: str = 'black') -> None:
        """
            Add a vector to the plot.

            Params:
                vector (Vector): vector to add
                origin (Point): origin of the vector
                color (str): color of the vector. Default is black

            Returns:
                None

            Raises:
                TypeError if vector is not a Vector
                ValueError if dimension is not equal or not origin given
        """
        if not isinstance(vector, Vector):
            raise TypeError('vector must be a Vector')
        if not isinstance(origin, Point):
            raise TypeError('origin must be a Point')
        if self.projection == '3d' and vector.dimension != 3:
            raise ValueError('Dimension must be 3 for 3d projection')
        if not origin:
            raise ValueError('not origin given')
        self.ax.quiver(origin.x, origin.y, origin.z, vector.get(
            'x'), vector.get('y'), vector.get('z'), color=color)

    def add_point(self, point: Point, color: str = 'black') -> None:
        """
            Add a point to the plot.

            Params:
                point (Point): point to add
                color (str): color of the point. Default is black

            Returns:
                None

            Raises:
                TypeError if point is not a Point
                ValueError if dimension is not equal or not point given
        """
        if not isinstance(point, Point):
            raise TypeError('point must be a Point')
        if self.projection == '3d' and point.dimension != 3:
            raise ValueError('Dimension must be 3 for 3d projection')
        if not point:
            raise ValueError('not point given')
        self.ax.scatter(point.x, point.y, point.z, color=color)

    def clear(self) -> None:
        """
            Clear the plot.

            Params:
                None

            Returns:
                None

            Raises:
                None
        """
        self.ax.clear()

    def show(self) -> None:
        """
            Show the plot.

            Params:
                None

            Returns:
                None

            Raises:
                None
        """
        plt.show()

    def save(self, path: str) -> None:
        """
            Save the plot.

            Params:
                path (str): path to save the plot

            Returns:
                None

            Raises:
                None
        """
        plt.savefig(path)



from .vector import Vector


class Point:
    """
        Point class
        Supported operators:
            +: add two points
            ==: check if two points are equal
    """

    def __init__(self, **kwargs):
        """
            Initialize a Point object.

            Params:
                x: x coordinate
                y: y coordinate
                z: z coordinate
                dimension: dimension of the vector. Default is 3

            Returns:
                A vector class instance.

            Raises:
                None
        """
        if kwargs is None:
            raise ValueError('No arguments given.')
        self.x = kwargs.get('x', 0)
        self.y = kwargs.get('y', 0)
        self.z = kwargs.get('z', 0)
        self.dimension = kwargs.get('dimension', 3)

    def midpoint(self, other: 'Point') -> 'Vector':
        """
            Calculate the midpoint between two points.

            Params:
                other: another point

            Returns:
                A vector class instance.

            Raises:
                None
        """

        if not isinstance(other, Point):
            raise TypeError('other must be a Point')
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        x = (self.x + other.x) / 2
        y = (self.y + other.y) / 2
        z = (self.z + other.z) / 2
        return Vector(x=x, y=y, z=z)

    def find_vector(self, other: 'Point') -> Vector:
        """
            Calculate the vector between two points.

            Params:
                other (Point): another point

            Returns:
                A vector class instance.

            Raises:
                TypeError if other is not a Point
                ValueError if dimensions are not equal
        """
        if not isinstance(other, Point):
            raise TypeError('other must be a Point')
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        return Vector(x=other.x - self.x, y=other.y - self.y, z=other.z - self.z)

    def __str__(self) -> str:
        """
            Return a string representation of the point.
        """
        return f'({self.x}, {self.y}, {self.z})'

    def __repr__(self) -> str:
        """
            Return a string representation of the point.
        """
        return f'Point({self.x}, {self.y}, {self.z})'

    def __eq__(self, other: 'Point') -> bool:
        """
            Check if two points are equal.

            Params:
                other (Point): another point

            Returns:
                True if points are equal, False otherwise.

            Raises:
                TypeError if other is not a Point
                ValueError if dimensions are not equal
        """
        if not isinstance(other, Point):
            raise TypeError('other must be a Point')
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __add__(self, other: 'Point') -> Vector:
        """
            Add two points.

            Params:
                other (Point): another point

            Returns:
                A vector class instance.

            Raises:
                TypeError if other is not a Point
                ValueError if dimensions are not equal
        """
        if not isinstance(other, Point):
            raise TypeError('other must be a Point')
        if self.dimension != other.dimension:
            raise ValueError('Dimensions must be equal')
        return Vector(x=self.x + other.x, y=self.y + other.y, z=self.z + other.z)
